/*
 * noVNC blue CSS
 * Copyright (C) 2012 Joel Martin
 * Copyright (C) 2013 Samuel Mannehed for Cendio AB
 * noVNC is licensed under the MPL 2.0 (see LICENSE.txt)
 * This file is licensed under the 2-Clause BSD license (see LICENSE.txt).
 */

.noVNC_panel {
  border:2px solid #fff;
  background:#04073d;
  color:#fff;
}

#noVNC_control_bar, #noVNC_control_bar_handle, .noVNC_panel .noVNC_heading {
  background: #080F80;
}
